using System;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Windows.Forms;
using System.ComponentModel;
using System.Runtime.InteropServices;

namespace RSSReader.Controls
{

	#region DragDrop event arguments classes 

	#region NodeStartDragEventArgs class 

	public class NodeStartDragEventArgs
	{

		#region Private variables 

		private TreeNode	node;
		private bool		cancel;

		#endregion

		#region Constructor 

		internal NodeStartDragEventArgs( TreeNode Node )
		{
			this.node = Node;
			cancel = false;
		}

		#endregion 

		#region Public properties 

		public TreeNode Node 
		{
			get { return node; }
		}

		public bool Cancel
		{
			get { return cancel; }
			set { cancel = value; }
		}

		#endregion 

	}

	#endregion 

	#region NodeDragOverEventArgs 

	public class NodeDragOverEventArgs
	{

		#region Private variables 

		private TreeNode		dragNode;
		private TreeNode		targetNode;
		private DragDropEffects	effect;

		#endregion 

		#region Constructor 

		public NodeDragOverEventArgs( TreeNode DragNode, TreeNode TargetNode )
		{
			this.dragNode = DragNode;
			this.targetNode = TargetNode;
			this.effect = DragDropEffects.All;
		}

		#endregion 

		#region Public properties 

		public TreeNode DragNode
		{
			get { return dragNode; }
		}

		public TreeNode TargetNode
		{
			get { return targetNode; }
		}

		public DragDropEffects Effect
		{
			get { return effect; }
			set { effect = value; }
		}

		#endregion 

	}

	#endregion 

	#region NodeDragDropEventArgs 

	public class NodeDragDropEventArgs
	{

		#region Private variables 

		private TreeNode		dragNode;
		private TreeNode		targetNode;

		#endregion 

		#region Constructor 

		public NodeDragDropEventArgs( TreeNode DragNode, TreeNode TargetNode )
		{
			this.dragNode = DragNode;
			this.targetNode = TargetNode;
		}

		#endregion 

		#region Public properties 

		public TreeNode DragNode
		{
			get { return dragNode; }
		}

		public TreeNode TargetNode
		{
			get { return targetNode; }
		}

		#endregion 

	}

	#endregion 

	#endregion 

	/// <summary>
	/// Enhances the intrinsic System.Windows.Forms.TreeView class by adding 
	/// such features as custom-drawn nodes and built-in drag-and-drop functionality.
	/// </summary>
	[
	Designer("System.Windows.Forms.Design.TreeViewDesigner, System.Design, Version=1.0.3300.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"),
	DefaultProperty("Nodes"),
	DefaultEvent("AfterSelect")
	]
	public class JSTree : TreeView 
	{

		#region Events 

		public delegate void NodeStartDragEventHandler( object sender, NodeStartDragEventArgs e );
		public delegate void NodeDragOverEventHandler( object sender, NodeDragOverEventArgs e );
		public delegate void NodeDragDropEventHandler( object sender, NodeDragDropEventArgs e );

		public event NodeStartDragEventHandler NodeStartDrag;
		public event NodeDragOverEventHandler NodeDragOver;
		public event NodeDragDropEventHandler NodeDragDrop;

		#endregion 

		#region Win32API support 

		[DllImport("User32.dll")]
		public static extern int GetUpdateRect( IntPtr hwnd, ref RECT rect, bool erase );

		[DllImport("User32.dll", SetLastError = true)]
		public static extern bool GetWindowRect( IntPtr handle, ref RECT rect );

		[DllImport("User32.dll")]
		public static extern IntPtr BeginPaint( IntPtr hWnd, ref PAINTSTRUCT paintStruct );

		[DllImport("User32.dll")]
		public static extern bool EndPaint( IntPtr hWnd, ref PAINTSTRUCT paintStruct );

		[StructLayout( LayoutKind.Sequential )]
			public struct RECT 
		{
			public int left;
			public int top;
			public int right;
			public int bottom;
		}

		[StructLayout( LayoutKind.Sequential ) ]
			public struct PAINTSTRUCT 
		{
			public IntPtr hdc;
			public int fErase;
			public RECT rcPaint;
			public int fRestore;
			public int fIncUpdate;
			public int Reserved1;
			public int Reserved2;
			public int Reserved3;
			public int Reserved4;
			public int Reserved5;
			public int Reserved6;
			public int Reserved7;
			public int Reserved8;
		}
	
		#endregion 

		#region Private variables 

		private Bitmap		bitmapInternal;
		private Graphics	graphicsInternal;
		private Form		dragForm;
		private PictureBox	dragPicture;

		private TreeNode	contextNode = null;
		private Image		dragImage = null;
		private bool		inDrag = false;
		private TreeNode	dragHilite = null;
		private Color		dragLastBackground;
		private Color		dragLastForeground;

		#endregion 

		#region Constructor 

		public JSTree() : base()
		{
			
			base.SetStyle( ControlStyles.DoubleBuffer, true );
			
			dragForm = new Form();
			dragForm.BackColor = Color.White;
			dragForm.AllowDrop = false;
			dragForm.ShowInTaskbar = false;
			dragForm.Opacity = 0.45;
			dragForm.ControlBox = false;
			dragForm.Size = new Size( 0, 0 );
			dragForm.Location = new Point( Screen.PrimaryScreen.WorkingArea.Width, 0 );
			dragForm.FormBorderStyle = FormBorderStyle.None;
			dragForm.TransparencyKey = this.BackColor;

			dragPicture = new PictureBox();
			dragPicture.SizeMode = PictureBoxSizeMode.AutoSize;
			dragPicture.Dock = DockStyle.Fill;
			dragForm.Controls.Add( dragPicture );

		}

		#endregion 

		#region Dispose 

		private void DisposeInternal() 
		{
		}

		/// <summary>Releases resources.</summary>
		/// <param name="disposing">true = Both managed and unmanaged, false = Unmanaged only.</param>
		protected override void Dispose(bool disposing) 
		{
		
			if( disposing )
				DisposeInternal();
			
			base.Dispose (disposing);

		}

		#endregion

		#region Public properties 

		/// <summary>
		/// Gets a value that indicates the TreeNode on which the last mouse
		/// action occured.  This allows the developer to tailor context menus
		/// and double-click actions to specific nodes.  If the last mouse 
		/// action did not occur on a TreeNode, this property will be NULL.
		/// </summary>
		[Browsable(false)]
		public TreeNode ContextNode
		{
			get { return contextNode; }
		}

		#endregion 

		#region Public methods 

		/// <summary>
		/// Retrieves the tree node at the point with the specified coordinates.  
		/// </summary>
		/// <param name="x">The System.Drawing.Point.X position to evaluate and retrieve the node from.</param>
		/// <param name="y">The System.Drawing.Point.Y position to evaluate and retrieve the node from.</param>
		/// <returns>The System.Windows.Forms.TreeNode at the specified location, in tree view coordinates.</returns>
		public new TreeNode GetNodeAt( int x, int y )
		{
			return GetNodeAt( new Point( x, y ) );
		}

		/// <summary>
		/// Retrieves the tree node at the point with the specified coordinates.  
		/// </summary>
		/// <param name="pt">The System.Drawing.Point to evaluate and retrieve the node from.</param>
		/// <returns>The System.Windows.Forms.TreeNode at the specified location, in tree view coordinates.</returns>
		public new TreeNode GetNodeAt( Point pt )
		{

			TreeNode loop = TopNode;
			while( loop != null && loop.IsVisible )
			{

				if( loop.IsVisible )
				{

					if( loop is IOwnerDrawNode )
					{
						if( ((IOwnerDrawNode) loop).NodeBounds.Contains( pt ) )
							return loop;
					}
					else
					{
						if( loop.Bounds.Contains( pt ) )
							return loop;
					}

				}

				loop = loop.NextVisibleNode;

			}

			return null;

		}

		#endregion 

		#region Custom Drag-n-Drop functionality 

		private void EndDrag()
		{

			if( !inDrag )
				return;

			inDrag = false;

			SetDragHighlight( null );
			dragForm.Visible = false;

		}

		protected override void OnItemDrag(ItemDragEventArgs e)
		{

			if( !(e.Item is TreeNode) )
				return;

			NodeStartDragEventArgs dragArgs = new NodeStartDragEventArgs( (TreeNode) e.Item );
			if( this.NodeStartDrag != null )
				this.NodeStartDrag( this, dragArgs );

			if( dragArgs.Cancel )
				return;
			
			dragImage = GetDragImage( (TreeNode) e.Item );

			this.inDrag = true;
			this.DoDragDrop( e.Item, DragDropEffects.Move );
			this.dragHilite = null;

		}

		private void RestoreDragHighlight()
		{
			dragHilite.BackColor = dragLastBackground;
			dragHilite.ForeColor = dragLastForeground;
		}

		private void SetDragHighlight( TreeNode target )
		{

			if( dragHilite != null && dragHilite != target )
				RestoreDragHighlight();

			if( dragHilite == target )
				return;

			dragHilite = target;

			if( target != null )
			{

				dragLastBackground = (target.BackColor != Color.Empty) ? target.BackColor : this.BackColor;
				dragLastForeground = (target.ForeColor != Color.Empty) ? target.ForeColor : this.ForeColor;
				
				target.BackColor = SystemColors.InactiveCaption;
				target.ForeColor = SystemColors.ActiveCaptionText;

			}

		}

		private bool IsNodeChildOf( TreeNode PossibleParent, TreeNode PossibleChild )
		{


			while( PossibleChild != null && PossibleChild != PossibleParent )
				PossibleChild = PossibleChild.Parent;

			return (PossibleChild == PossibleParent);

		}

		protected override void OnDragOver(DragEventArgs e)
		{

			if( !inDrag )
				return;

			Point mouseLocation = this.PointToClient( new Point( e.X, e.Y ) );
			TreeNode node = this.GetNodeAt( mouseLocation );

			if( node != null && !IsNodeChildOf( contextNode, node ) )
			{

				NodeDragOverEventArgs eventArgs = new NodeDragOverEventArgs( contextNode, node );
				if( this.NodeDragOver != null )
					this.NodeDragOver( this, eventArgs );

				e.Effect = eventArgs.Effect;

			}
			else
			{
				e.Effect = DragDropEffects.None;
			}

			SetDragHighlight( node );

		}

		protected override void OnDragLeave(EventArgs e)
		{

			base.OnDragLeave (e);
			RestoreDragHighlight();

			dragForm.Visible = false;

		}

		protected override void OnDragDrop(DragEventArgs e)
		{

			TreeNode target = dragHilite;
			
			base.OnDragDrop (e);
			RestoreDragHighlight();

			EndDrag();

			if( target != null && this.NodeDragDrop != null )
				this.NodeDragDrop( this, new NodeDragDropEventArgs( contextNode, target ) );

		}

		protected override void OnGiveFeedback(GiveFeedbackEventArgs e)
		{

			if( !inDrag )
				return;

			if( !dragForm.Visible && dragImage != null && this.ClientRectangle.Contains( this.PointToClient(Cursor.Position) ) )
			{
			
				dragPicture.Image = dragImage;

				dragForm.TopMost = true;
				dragForm.Visible = true;
				
				dragForm.Width = dragPicture.Width;
				dragForm.Height = dragPicture.Height;
				
				this.Focus();
			
			}

			Point mouseLocation = Cursor.Position;
			mouseLocation.Offset( 10, -5 );

			if( this.ClientRectangle.Contains( this.PointToClient( Cursor.Position ) ) )
				dragForm.Location = mouseLocation;
			else
			{
				dragForm.Visible = false;
				base.OnDragLeave( EventArgs.Empty );
				return;
			}

			#region See if we need to auto-scroll for the user 

			Point clientLocation = this.PointToClient( Cursor.Position );

			// Check for scroll down
			if( clientLocation.Y >= this.ClientRectangle.Height - this.ItemHeight )
			{
				
				TreeNode downNode = base.GetNodeAt( clientLocation );
				if( downNode != null && downNode.NextVisibleNode != null )
					downNode.NextVisibleNode.EnsureVisible();

			}
				// Check for scroll up
			else if( clientLocation.Y <= this.ItemHeight )
			{

				if( this.TopNode.PrevVisibleNode != null )
					this.TopNode.PrevVisibleNode.EnsureVisible();

			}

			#endregion 

		}

		private Image GetDragImage( TreeNode node )
		{

			Image img = null;

			if( node is IOwnerDrawNode )
				img = GetCustomDragImage( (IOwnerDrawNode) node );
			else
			{
				if( this.ImageList != null && node.ImageIndex >= 0 )
					img = ImageList.Images[ node.ImageIndex ];
			}

			if( img != null && img.Width > 255 )
				img = ClipImage( img, new Rectangle( 0, 0, 255, img.Height ) );

			return img;

		}

		private Image GetCustomDragImage( IOwnerDrawNode node )
		{

			Rectangle imageRect = this.ClientRectangle;

			Bitmap bmp = new Bitmap( imageRect.Width, imageRect.Height );
			Image img = Image.FromHbitmap( bmp.GetHbitmap( Color.White ) );
			Graphics g = Graphics.FromImage( img );
			g.Clear( this.BackColor );

			node.OnPaint( false, new PaintEventArgs( g, node.NodeBounds ) );

			Rectangle dragImageRect = node.NodeBounds;

			if( this.ImageList != null && ((TreeNode) node).ImageIndex >= 0 )
			{
				
				Image nodeIcon = ImageList.Images[ ((TreeNode) node).ImageIndex ];
				g.DrawImage( 
					nodeIcon,
					node.NodeBounds.Left - nodeIcon.Width - 5,
					node.NodeBounds.Top,
					nodeIcon.Width,
					nodeIcon.Height
					);

				dragImageRect = new Rectangle( 
					node.NodeBounds.Left - nodeIcon.Width - 5,
					node.NodeBounds.Top,
					node.NodeBounds.Width + nodeIcon.Width + 5,
					node.NodeBounds.Height 
					);

			}

			return ClipImage( img, dragImageRect );

		}

		private Image ClipImage( Image img, Rectangle ClipRect )
		{

			ClipRect.Intersect( this.ClientRectangle );

			Bitmap temp = new Bitmap( img );
			return temp.Clone( ClipRect, img.PixelFormat );

		}

		#endregion 

		#region Custom mouse processing 

		protected override void OnMouseMove(MouseEventArgs e)
		{

			// We do not get the OnMouseMove event while performing
			// a DragDrop operation, so it's safe to flag that we 
			// are no longer in a DragDrop operation if we were 
			// previously.
			if( inDrag )
				EndDrag();

			base.OnMouseMove (e);

		}

		protected override void OnMouseDown(MouseEventArgs e)
		{

			TreeNode hitTest = GetNodeAt( new Point( e.X, e.Y ) );
			contextNode = hitTest;

			if( hitTest != null )
				this.SelectedNode = hitTest;

			base.OnMouseDown( e );

		}

		#endregion 
 
		#region Custom paint functionality  

		protected override void OnResize(EventArgs e)
		{

			bitmapInternal = new Bitmap( this.ClientRectangle.Width, this.ClientRectangle.Height );
			graphicsInternal = Graphics.FromImage( bitmapInternal );

			base.OnResize (e);

		}

		/// <summary>Occurs when a Windows message is dispatched.</summary>
		/// <param name="message">Message to process.</param>
		/// <remarks>Overrides WM_PAINT, WM_ERASEBKGND.</remarks>
		protected override void WndProc(ref Message message) 
		{

			const int WM_PAINT = 0x000F;
			const int WM_PRINTCLIENT = 0x0318;
			const int WM_ERASEBKGND = 0x0014;
			const int WM_HSCROLL = 0x114;
			const int WM_VSCROLL = 0x115;

			switch( message.Msg ) 
			{

				case WM_ERASEBKGND:
					return;

				case WM_HSCROLL:
				case WM_VSCROLL:
					// There are some problems with the TreeView taking control of
					// our custom-paint nodes when scrolling, so we just invalidate
					// the entire TreeView when the user scrolls.  It's slower, 
					// but prettier ;)
					this.Invalidate( this.ClientRectangle, true );
					break;

				case WM_PAINT:
					RECT updateRect = new RECT();
					if( GetUpdateRect( message.HWnd, ref updateRect, false ) == 0 )
						break;

					PAINTSTRUCT paintStruct = new PAINTSTRUCT();
					IntPtr screenHdc = BeginPaint(message.HWnd, ref paintStruct);

					using( Graphics screenGraphics = Graphics.FromHdc( screenHdc ) ) 
					{

						IntPtr hdc = graphicsInternal.GetHdc();

						Message printClientMessage = Message.Create( Handle, WM_PRINTCLIENT, hdc, IntPtr.Zero );  
						DefWndProc( ref printClientMessage );

						graphicsInternal.ReleaseHdc( hdc );

						Rectangle clipRegion = Rectangle.FromLTRB(
							updateRect.left,
							updateRect.top,
							updateRect.right,
							updateRect.bottom 
							);

						OnPaint( new PaintEventArgs( graphicsInternal, clipRegion ) );

						screenGraphics.DrawImage( bitmapInternal, 0, 0 );

					}

					EndPaint( message.HWnd, ref paintStruct );

					return;

			}

			base.WndProc(ref message);

		}

		protected override void OnPaint(PaintEventArgs e)
		{

			Rectangle clipRect = new Rectangle(
				0,
				e.ClipRectangle.Top,
				this.ClientRectangle.Width,
				e.ClipRectangle.Height 
				);

			int visibleCount = this.Height / this.ItemHeight + 1;

			TreeNode loop = TopNode;
			for( int i = 0; i < visibleCount && loop != null; i++ )
			{

				if( loop is IOwnerDrawNode )
				{

					// Give the node a chance to perform custom painting.
					if( clipRect.IntersectsWith( ((IOwnerDrawNode) loop).NodeBounds ) )
						((IOwnerDrawNode) loop).OnPaint( loop == SelectedNode, e );

				}

				loop = loop.NextVisibleNode;

			}

		}

		#endregion 

	}

	#region IOwnerDrawNode interface 

	/// <summary>
	/// TreeNode objects which implement this interface will be 
	/// allowed a chance to perform custom drawing operations
	/// when the JSTree control is redrawn.
	/// </summary>
	public interface IOwnerDrawNode
	{

		void OnPaint( bool IsSelected, PaintEventArgs e );

		Rectangle NodeBounds { get; }

	}

	#endregion 

}

