using System;
using System.IO;
using System.Text;
using System.Xml;
using System.Collections;
using System.Windows.Forms;
using System.Threading;

using SharpTools;
using SharpTools.DOM;
using SharpTools.IDE;

using RSSReader.HTTP;
using RSSReader.RSS;

namespace RSSReader.RSS
{

	/// <summary>
	/// Maintains the cache for RSSFeed and related objects
	/// </summary>
	public class FeedManager
	{

		#region Delegate declarations 

		public delegate void RetrieveFeedAsyncCallback( RSSFeed feed, object Tag );

		#endregion 

		#region Private member variables 

		private static string		cachePath;
		private static Thread		asyncRequestThread;
		private static Queue		pendingRequests;
		private static bool			isPaused = false;
		private static bool			inDebugMode = false;
		private static bool			isRunning = true;

		#endregion 

		#region Constructor 

		static FeedManager()
		{

			string dataPath = Path.Combine( Environment.GetFolderPath( Environment.SpecialFolder.ApplicationData ), "STReader" );
			if( !Directory.Exists( dataPath ) )
				Directory.CreateDirectory( dataPath );

			cachePath = Path.Combine( dataPath, "cache" );
			if( !Directory.Exists( cachePath ) )
				Directory.CreateDirectory( cachePath );

			pendingRequests = new Queue();

			SharpTools.IDE.Debugger.GetInstance().EnterDesignMode += new SharpTools.IDE.Debugger.EnterDesignModeEventHandler(FeedManager_EnterDesignMode);
			SharpTools.IDE.Debugger.GetInstance().EnterRunMode += new SharpTools.IDE.Debugger.EnterRunModeEventHandler(FeedManager_EnterRunMode);

			asyncRequestThread = new Thread( new ThreadStart( AsyncRequestThreadWorker ) );
			asyncRequestThread.Name = "FeedManager Background Worker";
			asyncRequestThread.IsBackground = true;
			asyncRequestThread.Priority = ThreadPriority.BelowNormal;
			asyncRequestThread.Start();

		}

		#endregion 

		#region Public methods 

		public static void Stop()
		{

			try
			{

				setIsRunning( false );
			
				asyncRequestThread.Join( TimeSpan.FromSeconds( 5 ) );
				if( asyncRequestThread.IsAlive )
					asyncRequestThread.Abort();
		
				Application.DoEvents();

			}
			catch( Exception err )
			{
				System.Diagnostics.Debug.WriteLine( err.ToString() );
			}

		}

		public static void RetrieveFeedAsync( string url, object Tag, RetrieveFeedAsyncCallback Callback )
		{

			lock( pendingRequests )
			{

				if( !getIsRunning() )
					return;

				WebRequest request = new WebRequest();
				request.Url = url;
				
				AsyncWebRequest asyncRequest = new AsyncWebRequest( request, Tag, Callback );
				pendingRequests.Enqueue( asyncRequest );

			}

		}

		public static bool IsFeedCached( string url )
		{

			string cachedFile = GetCacheFilename( url );
			return File.Exists( cachedFile );

		}

		public static RSSFeed GetCachedFeed( string url )
		{

			try
			{

				string cachedFile = GetCacheFilename( url );
				if( !File.Exists( cachedFile ) )
					return null;

				string xmlData = "";
				using( StreamReader reader = File.OpenText( cachedFile ) )
				{
					xmlData = reader.ReadToEnd();
				}

				if( xmlData == "" )
				{
					File.Delete( cachedFile );
					return null;
				}

				XmlDocument cachedFeedXml = new XmlDocument();
				cachedFeedXml.LoadXml( xmlData );

				RSSFeed feed = new RSSFeed();
				feed.ReadFromXml( cachedFeedXml.DocumentElement, url );

				return feed;

			}
			catch( Exception err )
			{
				System.Diagnostics.Debug.WriteLine( err.ToString() );
				return null;
			}

		}

		public static RSSFeed RetrieveFeed( string url )
		{

			try
			{
				
				if( !getIsRunning() )
					return null;

				setIsPaused( true );

				RSSFeed cachedFeed = GetCachedFeed( url );
				if( cachedFeed != null )
					return cachedFeed;

				WebRequest request = new WebRequest();
				request.LastModified = DateTime.Today.Subtract( TimeSpan.FromDays( 30 ) );
				request.Url = url;

				WebResponse response = request.GetResponse();
				RSSFeed newFeed = GetFeedFromResponse( response );

				if( newFeed != null )
					SaveFeed( newFeed );

				return newFeed;

			}
			finally
			{
				setIsPaused( false );
			}

		}

		public static void SaveFeed( RSSFeed feed )
		{

			StringBuilder buffer = new StringBuilder();
			XmlTextWriter writer = XmlHelper.CreateXmlWriter( new StringWriter( buffer ) );

			feed.WriteToXml( writer );

			string filename = GetCacheFilename( feed.XmlUrl );
			if( File.Exists( filename ) )
				File.Delete( filename );

			using( StreamWriter fileOut = File.CreateText( filename ) )
			{
				fileOut.WriteLine( buffer.ToString() );
			}

		}

		public static void DeleteFeedFromCache( string feedUrl )
		{

			RSSFeed cached = GetCachedFeed( feedUrl );
			if( cached != null )
			{

				foreach( RSSItem item in cached.Items )
					if( item.CommentFeedUrl != "" )
						DeleteFeedFromCache( item.CommentFeedUrl );

			}

			string filename = GetCacheFilename( feedUrl );
			if( File.Exists( filename ) )
				File.Delete( filename );

		}

		#endregion

		#region Private utility methods 

		private static bool getInDebugMode()
		{
			lock( typeof(FeedManager) )
			{
				return inDebugMode;
			}
		}

		private static void setInDebugMode( bool value )
		{
			lock( typeof(FeedManager) )
			{
				inDebugMode = value;
			}
		}

		private static bool getIsPaused()
		{
			lock( typeof(FeedManager) )
			{
				return isPaused;
			}
		}

		private static void setIsPaused( bool value )
		{
			lock( typeof(FeedManager) )
			{
				isPaused = value;
			}
		}

		private static bool getIsRunning()
		{
			lock( typeof(FeedManager) )
			{
				return isRunning;
			}
		}

		private static void setIsRunning( bool value )
		{
			lock( typeof(FeedManager) )
			{
				isRunning = value;
			}	
		}

		private static RSSFeed GetFeedFromResponse( WebResponse response )
		{

			try
			{

				if( response == null || response.StatusCode != System.Net.HttpStatusCode.OK )
					return null;

				string xmlData = response.Data;
				if( xmlData == "" )
					return null;

				XmlDocument newFeedXml = new XmlDocument();
				newFeedXml.LoadXml( xmlData );

				RSSFeed newFeed = new RSSFeed();
				newFeed.ReadFromXml( newFeedXml.DocumentElement, response.Url );

				newFeed.LastModified = response.LastModified;
				newFeed.ETag = response.ETag;
				newFeed.LastRefresh = DateTime.Now;

				foreach( RSSItem item in newFeed.Items )
					item.RetrieveComments();

				return newFeed;

			}
			catch( Exception err )
			{
				System.Diagnostics.Debug.WriteLine( err.ToString() );
				return null;
			}

		}

		private static string GetCacheFilename( string url )
		{

			if( url.ToLower().StartsWith( "http://" ) )
				url = url.Substring( 7 );
			else if( url.ToLower().StartsWith( "file://" ) )
			{
				url = url.Substring( 7 );
				url = Path.GetFileNameWithoutExtension( url );
			}

			string result = "";
			foreach( char ch in url.ToCharArray() )
			{
				
				if( char.IsLetterOrDigit( ch ) || ch == '.' || ch == '-' || ch == '_' )
					result += ch;
				else
					result += "-";

			}

			return Path.Combine( cachePath, result + "-feed.xml" );

		}

		#endregion 

		#region Async Request Thread 

		private static void AsyncRequestThreadWorker()
		{

			try
			{

				while( true )
				{

					if( !getIsRunning() )
						return;

					AsyncWebRequest request = null;

					// We really want this thread to be secondary to all other
					// threads that are currently running in the IDE and in 
					// the user's application.
					Application.DoEvents();
					Thread.Sleep( 100 );
					
					// If the async thread is paused or the user is in debug (run) 
					// mode, do not process async requests.
					if( getIsPaused() || getInDebugMode() || !getIsRunning() )
						continue;

					lock( pendingRequests )
					{

						if( pendingRequests.Count > 0 )
						{
							request = (AsyncWebRequest) pendingRequests.Dequeue();
						}

					}

					if( request != null && getIsRunning() && !getIsPaused() && !getInDebugMode() )
						request.PerformRequest();

				}

			}
			catch( ThreadAbortException )
			{
			}
			catch( Exception err )
			{
				System.Diagnostics.Debug.WriteLine( "ERROR IN FEEDMANAGER THREAD! " + err.ToString() );
			}

		}

		#endregion 

		#region Debugger events 

		private static void FeedManager_EnterRunMode(SharpTools.IDE.Debugger.DebugModeChangedEventArgs args)
		{
			setInDebugMode( true );
		}

		private static void FeedManager_EnterDesignMode(SharpTools.IDE.Debugger.DebugModeChangedEventArgs args)
		{
			setInDebugMode( false );
		}

		#endregion 

		#region Private classes 

		private class AsyncWebRequest
		{

			#region Public properties 

			public RetrieveFeedAsyncCallback Callback;

			public object		Tag;
			public WebRequest	Request;

			#endregion 

			#region Constructor 

			private AsyncWebRequest() {}

			public AsyncWebRequest( WebRequest Request, object Tag, RetrieveFeedAsyncCallback Callback )
			{
				this.Request = Request;
				this.Tag = Tag;
				this.Callback = Callback;
			}

			#endregion 

			#region Public methods 

			public void PerformRequest()
			{

				try
				{

					WebResponse response = Request.GetResponse();
					RSSFeed feed = GetFeedFromResponse( response );

					if( Callback != null )
						Callback( feed, Tag );

				}
				catch( Exception err )
				{
				}

			}

			#endregion 

		}

		#endregion

	}

}
