using System;
using System.Xml; 
using System.Xml.Schema;
using System.IO; 
using System.Collections;
using System.Xml.Serialization;
using System.Runtime.CompilerServices;
using System.Threading;
using System.Xml.Xsl;
using System.Net;
using System.Xml.XPath;
using System.Diagnostics;
using System.Text;
using System.ComponentModel;

namespace RSSReader.RSS
{

	/// <summary>
	/// Represents an RSS subscription feed
	/// </summary>
	[DefaultProperty("Title")]
	public class RSSFeed
	{

		#region Private member variables 

		private XmlNamespaceManager ns;
		private XmlNode				sourceNode;
		private RSSItemCollection	items;
		private RSSFeedImage		image;

		private string			title;
		private string			xmlUrl;
		private string			htmlUrl;
		private string			language;
		private string			description;
		private bool			alertNewItems;
		private DateTime		lastRefresh;
		private TimeSpan		refreshRate;
		private TimeSpan		maxItemAge;
		private int				maxHistorySize;
		private bool			purgeUnreadItems;

		private string			lastModified;
		private string			eTag;

		#endregion 

		#region Constructor 

		public RSSFeed()
		{
			
			items = new RSSItemCollection();

			#region Set default property values 

			title = "";
			xmlUrl = "";
			htmlUrl = "";
			language = "en-US";
			description = "";
			alertNewItems = true;
			lastRefresh = DateTime.Now;
			refreshRate = TimeSpan.FromHours( 4 );
			maxItemAge = TimeSpan.FromDays( 120 );
			purgeUnreadItems = true;
			lastModified = DateTime.Now.ToString();
			eTag = "";

			#endregion 

		}

		#endregion 

		#region Public properties 

		[Category("Feed"),Description("The name of this feed")]
		public string Title
		{
			get { return title; }
			set { title = value; }
		}

		[Category("Feed"),Description("The description assigned to this feed")]
		[DefaultValue("")]
		[ReadOnly(true)]
		public string Description
		{
			get { return description; }
			set { description = value; }
		}

		[Category("Address"),Description("The address of the XML feed")]
		[DefaultValue("")]
		public string XmlUrl
		{
			get { return xmlUrl; }
			set { xmlUrl = value; }
		}

		[Category("Address"),Description("The address (if available) of the website hosting the XML feed")]
		[DefaultValue("")]
		public string HtmlUrl
		{
			get { return htmlUrl; }
			set { htmlUrl = value; }
		}

		[Browsable(false)]
		public RSSFeedImage Image
		{
			get { return image; }
			set { image = value; }
		}

		[Category("Options"),Description("Indicates whether to show a pop-up notice when new items are added to the feed")]
		[DefaultValue(true)]
		public bool AlertOnNewItems
		{
			get { return alertNewItems; }
			set { alertNewItems = value; }
		}

		[Category("Options"),Description("How often to check for updates to this feed")]
		public TimeSpan RefreshRate
		{
			get { return refreshRate; }
			set { refreshRate = value; }
		}

		[Category("Housekeeping"),Description("The maximum amount of time to keep items in the history.  After this period has passed, the item will be deleted")]
		public TimeSpan MaxItemAge
		{
			get { return maxItemAge; }
			set { maxItemAge = value; }
		}

		[Category("Housekeeping"),Description("The maximum number of items to keep for this feed")]
		[DefaultValue(50)]
		public int MaxItemCount
		{
			get { return maxHistorySize; }
			set { maxHistorySize = value; }
		}

		[Category("Housekeeping"),Description("Indicates whether unread items will be purged.  If set to FALSE, unread items will not be deleted even when they pass the MaxItemAge limit")]
		[DefaultValue(true)]
		public bool PurgeUnreadItems
		{
			get { return purgeUnreadItems; }
			set { purgeUnreadItems = value; }
		}

		[Category("Info"),Description("The last time this feed was refreshed")]
		[ReadOnly(true)]
		public DateTime LastRefresh
		{
			get { return lastRefresh; }
			set { lastRefresh = value; }
		}
		
		[Category("Info"),Description("The last time this feed was updated by the publisher")]
		[ReadOnly(true)]
		public string LastModified
		{
			get { return lastModified; }
			set { lastModified = value; }
		}
		
		[Browsable(false)]
		public string ETag
		{
			get { return eTag; }
			set { eTag = value; }
		}

		[Category("Info"),Description("The language in which the feed was created")]
		[ReadOnly(true)]
		[DefaultValue("en-us")]
		public string Language
		{
			get { return language; }
			set { language = value; }
		}

		[Browsable(false)]
		public RSSItemCollection Items
		{
			get { return items; }
		}

		#endregion 

		#region Public methods
 
		public int GetItemCount()
		{
			
			int count = 0;
			foreach( RSSItem item in this.items )
			{

				if( !item.Deleted )
					count += 1;

			}

			return count;

		}

		public int GetItemsReadCount()
		{

			int markedAsRead = 0;
			foreach( RSSItem item in items )
			{
				if( item.MarkedAsRead )
				{
					markedAsRead += 1;
				}
			}

			return markedAsRead;

		}

		private PopupLink[] GetPopupLinks( ArrayList items )
		{

			PopupLink[] links = new PopupLink[ items.Count ];
			for( int i = 0; i < links.Length; i++ )
			{

				RSSItem item = (RSSItem) items[i];
				links[i] = new PopupLink( item.Title, item.Guid );

			}

			return links;

		}

		public void Merge( RSSFeed newFeed )
		{

			this.lastModified = newFeed.lastModified;
			this.eTag = newFeed.eTag;
			this.lastRefresh = newFeed.LastRefresh;

			ArrayList newItems = new ArrayList();

			foreach( RSSItem newItem in newFeed.Items )
			{

				if( !this.ContainsItem( newItem ) )
				{

					// Don't bother with items too old to keep
					if( newItem.DatePublished.Add( this.MaxItemAge ) < DateTime.Today )
						continue;

					newItem.Parent = this;
					newItems.Add( newItem );

					// I know this seems strange, but since nearly all feed items simply
					// have a date with no time component, adding an insignificant
					// time component will allow us to sort them by date in a manner 
					// that makes sense to the user.  In other words, to sort the
					// list of items by date in descending order will always show the
					// most recently added items first, instead of last.
					newItem.DatePublished = newItem.DatePublished.AddMilliseconds( this.items.Count * 2 );

					this.items.Insert( 0, newItem );

				}
				else
				{
					ReplaceItem( newItem );	
				}

			}

			if( this.alertNewItems && newItems.Count > 0 )
				PopupManager.ShowPopup( "RSS Reader", this.Title, GetPopupLinks( newItems ) );

		}

		public bool ContainsItem( RSSItem item )
		{

			foreach( RSSItem child in items )
			{	
				
				if( child.Guid.ToLower() == item.Guid.ToLower() || child.Link.ToLower() == item.Link.ToLower() )
					return true;

			}

			return false;

		}

		private void ReplaceItem( RSSItem item )
		{

			foreach( RSSItem child in items )
			{

				if( child.Guid.ToLower() == item.Guid.ToLower() || child.Link.ToLower() == item.Link.ToLower() )
				{

					if( item.DatePublished <= child.DatePublished || child.Deleted )
						return;

					// Keep some important current properties 
					item.DatePublished = child.DatePublished;
					item.MarkedAsRead = child.MarkedAsRead;
					item.MarkedForFollowUp = child.MarkedForFollowUp;
					item.SortField = child.SortField;
					item.SortOrder = child.SortOrder;
					
					this.items.Remove( child );

					break;

				}

			}

			item.Parent = this;
			items.Add( item );

		}

		public RSSItem GetOldestItem()
		{

			RSSItem oldest = null;
			foreach( RSSItem item in Items )
			{

				if( item.Deleted )
					continue;

				if( oldest == null || item.DatePublished < oldest.DatePublished )
					oldest = item;

			}

			return oldest;

		}

		public void Sort( ItemSortField SortField, ItemSortOrder SortOrder )
		{
			items.Sort( SortField, SortOrder );
		}

		#endregion 

		#region Serialization 

		internal void WriteToXml( XmlTextWriter writer )
		{

			writer.WriteStartElement( "rss" );
			writer.WriteAttributeString( "version", "2.0" );

			writer.WriteStartElement( "channel" );

			XmlHelper.WriteSimpleElement( writer, "title", this.title );
			XmlHelper.WriteSimpleElement( writer, "link", this.htmlUrl );
			XmlHelper.WriteSimpleElement( writer, "description", this.description );
			XmlHelper.WriteSimpleElement( writer, "language", this.language );

			XmlHelper.WriteSimpleElement( writer, "lastModified", this.lastModified );
			XmlHelper.WriteSimpleElement( writer, "eTag", this.eTag );
			XmlHelper.WriteSimpleElement( writer, "lastRefresh", this.lastRefresh.ToString() );
			XmlHelper.WriteSimpleElement( writer, "refreshRate", this.refreshRate.ToString() );
			XmlHelper.WriteSimpleElement( writer, "maxItemAge", this.maxItemAge.ToString() );
			XmlHelper.WriteSimpleElement( writer, "maxHistorySize", this.maxHistorySize.ToString() );
			XmlHelper.WriteSimpleElement( writer, "purgeUnreadItems", this.purgeUnreadItems.ToString() );
			XmlHelper.WriteSimpleElement( writer, "alertNewItems", this.alertNewItems.ToString() );

			if( image != null )
				image.WriteToXml( writer );

			// Implementation of MaxItemAge
			while( this.GetItemCount() > 0 )
			{

				// Find the oldest item in the feed
				RSSItem oldest = GetOldestItem();
				if( oldest == null )
					break;

				// Delete the oldest item if it is older than MaxItemAge.
				// Note that this also deletes its associated comment feeds 
				// from the cache.
				if( oldest.DatePublished.Add( this.MaxItemAge ) < DateTime.Today )
					oldest.Deleted = true;
				else
					break;

			}

			// Implementation of MaxItemCount
			while( this.GetItemCount() > this.MaxItemCount )
			{
				
				RSSItem oldestItem = GetOldestItem();
				if( oldestItem == null )
					break;

				// Delete the item.  Note that this also deletes its associated
				// comment feeds from the cache.
				oldestItem.Deleted = true;

			}

			foreach( RSSItem item in items )
			{
				item.WriteToXml( writer );
			}

			writer.WriteEndElement(); // </channel>

			writer.WriteEndElement(); // </rss>

		}

		internal void ReadFromXml( XmlNode Feed, string FeedUrl )
		{

			// Set up commonly used RSS namespaces
			ns = new XmlNamespaceManager( Feed.OwnerDocument.NameTable );
			ns.AddNamespace( "dc", "http://purl.org/dc/elements/1.1/" );
			ns.AddNamespace( "content", "http://purl.org/rss/1.0/modules/content/" );
			ns.AddNamespace( "wfw", "http://wellformedweb.org/CommentAPI/" );
			ns.AddNamespace( "rss", GetDefaultNamespace( Feed ) );

			sourceNode = Feed.SelectSingleNode( "//rss:channel", ns );

			// Read common and simple feed properties 
			title = GetValue( "", "rss:title", "title" );
			htmlUrl = GetValue( "", "rss:link", "link" );
			language = GetValue( "en-us", "rss:language", "dc:language" );
			description = GetValue( "", "rss:description", "description" );
			xmlUrl = FeedUrl;

			// Read custom properties, if stored in this xml file
			lastRefresh = GetValue( new DateTime( 1970, 1, 1 ), "lastRefresh" );
			lastModified = GetValue( "", "lastModified" );
			eTag = GetValue( "", "eTag" );
			maxItemAge = GetValue( TimeSpan.FromDays( 120 ), "maxItemAge" );
			maxHistorySize = GetValue( 100, "maxHistorySize" );
			purgeUnreadItems = GetValue( true, "purgeUnreadItems" );
			alertNewItems = GetValue( true, "alertNewItems" );

			// Retrieve the refresh rate, or the time between automatic refreshes.
			TimeSpan defaultRefreshRate = TimeSpan.FromHours(8);
			int ttl = GetValue( 0, "ttl" );
			if( ttl > 0 )
				refreshRate = TimeSpan.FromMinutes( ttl );
			else
				refreshRate = (TimeSpan) GetValue( defaultRefreshRate, "refreshRate" );

			GetImageInfo();

			// Read individual items from feed
			XmlNodeList xmlItems = Feed.SelectNodes( "//rss:item", ns );
			foreach( XmlNode xmlItem in xmlItems )
			{

				RSSItem item = new RSSItem( xmlItem );
				item.Parent = this;
				items.Add( item );

			}

		}

		private void GetImageInfo()
		{

			XmlNode imageNode = sourceNode.SelectSingleNode( "rss:image", ns );
			if( imageNode == null )
				image = null;
			else
				image = new RSSFeedImage( imageNode );

		}

		private string GetDefaultNamespace( XmlNode Feed )
		{

			XmlDocument doc = Feed.OwnerDocument;
			XmlNode docElement = doc.DocumentElement;

			string rssNamespaceUri = String.Empty; 

			// RSS 0.9 or 1.0
			if( docElement.LocalName.Equals("RDF") && docElement.NamespaceURI.Equals("http://www.w3.org/1999/02/22-rdf-syntax-ns#"))
			{ 
				
				// Figure out if 0.9 or 1.0 by peeking at namespace URI of <channel> element 
				if( (docElement.ChildNodes.Count > 0) && docElement.FirstChild.Name.Equals("channel") )
					rssNamespaceUri = docElement.FirstChild.NamespaceURI;
				else 
					// No channel, just assume RSS 1.0 
					rssNamespaceUri = "http://purl.org/rss/1.0/";

			}
			else if( docElement.LocalName.Equals("rss") )
			{ 

				// RSS 0.91 or 2.0 
				rssNamespaceUri = docElement.NamespaceURI; 								
			}
			else
			{
				throw new ApplicationException("This XML document does not look like an RSS feed.  Document element: " + docElement.Name );
			}

			return rssNamespaceUri;

		}

		private string GetValue( XmlNode Node, string DefaultValue, params string[] ElementNames )
		{

			if( ElementNames == null || ElementNames.Length == 0 )
				throw new ArgumentException( "You must provide an Element Name or a list of Element Names", "ElementNames" );

			foreach( string elementName in ElementNames )
			{

				XmlNode node = Node.SelectSingleNode( elementName, ns );
				if( node != null )
					return node.InnerText;

			}

			return DefaultValue;

		}

		private int GetValue( int DefaultValue, params string[] ElementNames )
		{
			
			string intValue = GetValue( DefaultValue.ToString(), ElementNames );
			return int.Parse( intValue );

		}

		private TimeSpan GetValue( TimeSpan DefaultValue, params string[] ElementNames )
		{
			
			string spanValue = GetValue( DefaultValue.ToString(), ElementNames );
			return TimeSpan.Parse( spanValue );

		}

		private DateTime GetValue( DateTime DefaultValue, params string[] ElementNames )
		{
			
			string dateValue = GetValue( DefaultValue.ToString(), ElementNames );
			return DateTime.Parse( dateValue );

		}

		private bool GetValue( bool DefaultValue, params string[] ElementNames )
		{
			
			string boolValue = GetValue( DefaultValue.ToString(), ElementNames );
			return bool.Parse( boolValue );

		}

		private string GetValue( string DefaultValue, params string[] ElementNames )
		{
			return GetValue( sourceNode, DefaultValue, ElementNames );
		}

		#endregion 

	}


}
