using System;
using System.IO;
using System.Xml;
using System.Xml.Xsl;
using System.Text;
using System.Drawing;
using System.Reflection;
using System.Collections;
using System.ComponentModel;
using System.Windows.Forms;
using System.Data;
using System.Threading;
using System.Runtime;
using System.Runtime.InteropServices;
using Microsoft.Win32;

using RSSReader;
using RSSReader.OPML;
using RSSReader.Controls;
using RSSReader.RSS;

using SharpTools;
using SharpTools.PluginInterface;
using SharpTools.IDE;

namespace RSSReaderPlugin
{

	/// <summary>
	/// Implements a plugin that provides an RSS Aggregator 
	/// for the Visual Studio .NET development environment.
	/// </summary>
	public class Plugin : ISharpToolsPlugin
	{

		#region Constants 

		private const string EXPLORER_TOOLWINDOW = "{EF062910-1290-46e8-8560-1F4288DB1F28}";
		private const string FEEDLIST_TOOLWINDOW = "{1FC57319-4FCA-4560-841F-29A85F9FFE5F}";
		private const string SEARCHRESULTS_TOOLWINDOW = "{586B81FE-623A-4c23-83E9-720E99D7DBB2}";

		#endregion 

		#region Global variables 

		private static EnvDTE._DTE dte;
		public static EnvDTE._DTE VSIDE
		{
			get { return dte; }
		}

		#endregion 

		#region Private variables 

		private Toolbar					toolbar;
		private ToolbarButton			button;
		private SubscriptionExplorer	tvwSubscriptions;
		private FeedListView			feedView;
		private SearchResultsView		searchResults;
		private Toolwindow				feedViewWindow;
		private Toolwindow				subscriptionsWindow;
		private Toolwindow				searchResultsWindow;
		private EnvDTE.Window			browserWindow;

		#endregion 

		#region Constructor 

		public Plugin()
		{
			// Empty because it's not used.  This constructor
			// is only here because reflection needs a default
			// constructor to instantiate objects.
		}

		#endregion 

		#region ISharpToolsPlugin Members 

		public string Description
		{
			get { return "Provides an RSS Aggregator integrated with the VS.NET IDE"; }
		}

		public string ID
		{
			get { return "SharpTools.Plugins.RSSReader"; }
		}

		public string Name
		{
			get { return "RSS Reader Plugin"; }
		}

		public void OnConnect()
		{

			EnvDTE.Window baseWindow = (EnvDTE.Window) SharpTools.IDE.CommandWindow.GetInstance().Object;
			dte = (EnvDTE._DTE) baseWindow.Collection.Parent;
			
			InitToolbar();

			PopupManager.SetFormSize( new Size( 250, 120 ) );

		}

		public void OnDisconnect(SharpTools.PluginInterface.DisconnectMode disconnectMode)
		{

			try
			{

				if( this.feedViewWindow != null )
					feedViewWindow.Close( SharpTools.DOM.SaveMode.No );

				tvwSubscriptions.Stop();
				PopupManager.Stop();
				FeedManager.Stop();

			}
			catch( Exception err )
			{
				System.Diagnostics.Debug.WriteLine( err.ToString() );
			}

		}

		#endregion

		#region Private methods 

		private void InitToolbar()
		{

			toolbar = new Toolbar( "RSS Reader Plugin" );
			button = (ToolbarButton) toolbar.Controls.Add( ControlType.Button, "RSSReader" );
			button.Caption = "RSS";
			button.Picture = 2569;
			button.Tooltip = "Show the SharpTools RSS Reader window";
			button.ButtonStyle = ButtonStyle.IconAndCaption;
			button.Click += new EventHandler(button_Click);
		}

		private Image GetToolwindowIcon()
		{

			Type type = this.GetType();
			string ns = type.Namespace;
			string imageFile = ns + ".MSINC.gif";
			using( Stream stream = type.Assembly.GetManifestResourceStream( imageFile ) )
			{
				return Image.FromStream( stream );
			}

		}

		private void InitExplorer()
		{

			tvwSubscriptions = new SubscriptionExplorer();
			tvwSubscriptions.NodeDoubleClick += new RSSReader.SubscriptionExplorer.NodeDoubleClickHandler(tvwSubscriptions_NodeDoubleClick);
			tvwSubscriptions.NodeFeedRefreshed += new RSSReader.SubscriptionExplorer.NodeFeedRefreshedHandler(tvwSubscriptions_NodeFeedRefreshed);
			tvwSubscriptions.SearchComplete += new RSSReader.SubscriptionExplorer.SearchCompletedHandler(tvwSubscriptions_SearchComplete);

			subscriptionsWindow = new Toolwindow( 
				new Guid( EXPLORER_TOOLWINDOW ), 
				"RSS Reader - Subscribed Feeds", 
				tvwSubscriptions 
				);

			subscriptionsWindow.Visible = true;
			subscriptionsWindow.SetTabPicture( this.GetToolwindowIcon() );

			try
			{
				// Try to size the toolwindow.  If it does not start out docked, 
				// this will work fine.  If, however, it does come up docked, 
				// and exception will be thrown that we just ignore.
				Application.DoEvents();
				subscriptionsWindow.Width = 200;
				subscriptionsWindow.Height = 500;
			}
			catch( Exception ) {}

			tvwSubscriptions.Refresh();

			System.Threading.Thread loadSubscriptionsWorker = new Thread( new ThreadStart( this.LoadSubscriptions ) );
			loadSubscriptionsWorker.IsBackground = true;
			loadSubscriptionsWorker.Priority = ThreadPriority.BelowNormal;
			loadSubscriptionsWorker.Start();

		}

		private void LoadSubscriptions()
		{
			
			// Give the SharpTools splash screen some time to go away, and the user
			// a chance to dismiss it if they wish.
			Thread.Sleep( TimeSpan.FromSeconds( 5 ) );

			tvwSubscriptions.SubscriptionFile = Path.Combine( Globals.InstallPath, "RSSReader.opml" );
			tvwSubscriptions.BeginInvoke( new MethodInvoker( tvwSubscriptions.RefreshTree ) );

		}

		private void ShowFeedList( SubscriptionTreeNode node )
		{

			if( feedViewWindow == null )
			{

				feedView = new FeedListView();
				feedView.FeedItemChanged += new RSSReader.Controls.FeedListView.FeedItemChangedHandler(FeedView_FeedItemChanged);
				feedView.FeedItemSelected += new RSSReader.Controls.FeedListView.FeedItemSelectedHandler(FeedView_FeedItemSelected);
				
				feedViewWindow = new Toolwindow( new Guid( FEEDLIST_TOOLWINDOW ), "RSS Reader", feedView );
				feedViewWindow.Visible = true;
				feedViewWindow.Linkable = true;
				feedViewWindow.IsFloating = false;
				feedViewWindow.SetTabPicture( this.GetToolwindowIcon() );

			}

			feedViewWindow.Visible = true;
			feedViewWindow.Activate();
			feedViewWindow.Caption = "RSS Reader - " + node.Subscription.Title;

			feedView.Subscription = node.Subscription;

		}

		#endregion

		private void tvwSubscriptions_NodeDoubleClick(object sender, OPMLNode node)
		{

			if( tvwSubscriptions.SelectedNode is SubscriptionTreeNode )
				ShowFeedList( (SubscriptionTreeNode) tvwSubscriptions.SelectedNode );

		}

		private string BuildItemPreview( RSSItem item )
		{

			StringBuilder itemXmlBuffer = new StringBuilder();
			XmlTextWriter itemXmlWriter = XmlHelper.CreateXmlWriter( new StringWriter( itemXmlBuffer ) );
			item.WriteToXml( itemXmlWriter );

			System.Type thisType = this.GetType();
			Assembly	thisAssembly = thisType.Assembly;
			string		thisNS = thisType.Namespace;
			string		templateName = thisNS + ".XSLT." + "RSSItemTemplate.xslt";

			XsltArgumentList args = new XsltArgumentList();
			args.AddParam( "FeedUrl", "", item.Parent.HtmlUrl );
			args.AddParam( "FeedName", "", item.Parent.Title );
			args.AddParam( "ItemUrl", "", item.Link );

			args.AddParam( "CommentUrl", "", item.CommentPostUrl );
			
			if( item.Comments != null )
				args.AddParam( "CommentCount", "", item.Comments.Items.Count );
			else
				args.AddParam( "CommentCount", "", "0" );

			if( item.Parent.Image != null )
				args.AddParam( "ImageUrl", "", item.Parent.Image.Url );

			using( StreamReader xsltReader = new StreamReader( thisAssembly.GetManifestResourceStream( templateName ) ) )
			{

				string itemHtml = XmlHelper.RenderXML( 
					new StringReader( itemXmlBuffer.ToString() ),
					xsltReader,
					args
					);

				string result = System.Web.HttpUtility.HtmlDecode( itemHtml );
				result = result.Replace( "", "'" );
				result = result.Replace( "", "'" );
				result = result.Replace( "", "\"" );
				result = result.Replace( "", "\"" );
				result = result.Replace( "", "..." );
				result = result.Replace( "", "-" );

				return result;

			}
				
		}

		private string WritePreviewToFile( string htmlData )
		{

			lock(this)
			{

				string filename = Path.Combine( Globals.InstallPath, "RSSReaderPreview.htm" );
				if( File.Exists( filename ) )
					File.Delete( filename );

				using( StreamWriter writer = File.CreateText ( filename ) )
				{
					writer.WriteLine( htmlData );
				}

				return "file://" + filename;

			}

		}

		private void FeedView_FeedItemSelected(object sender, RSSItem item)
		{

			string preview = BuildItemPreview( item );
			string filename = WritePreviewToFile( preview );

			EnvDTE.Window browser = VSIDE.ItemOperations.Navigate( filename, EnvDTE.vsNavigateOptions.vsNavigateOptionsDefault );
			if( browser != this.browserWindow )
				AttachBrowserEvents( browser );

			TreeNode feedNode = tvwSubscriptions.FindNodeForFeed( item.Parent.XmlUrl );
			if( feedNode != null && feedNode != tvwSubscriptions.SelectedNode )
			{
				tvwSubscriptions.SelectedNode = (OPMLTreeNode) feedNode;
				feedNode.Parent.EnsureVisible();
				feedNode.EnsureVisible();
			}

		}

		private void AttachBrowserEvents( EnvDTE.Window browser )
		{

			this.browserWindow = browser;

			// TODO: Get the IHTMLDocument reference to the web browser document
			// and attach the NavigateComplete event

		}

		private void tvwSubscriptions_NodeFeedRefreshed(object sender, OPMLItemNode node)
		{

			if( feedView == null )
				return;

			if( feedView.Subscription.XmlUrl.ToLower() == ((OPMLItemNode) node).XmlUrl.ToLower() )
				feedView.Subscription = node;

		}

		private void tvwSubscriptions_SearchComplete(object sender, FeedSearchResults results)
		{

			if( searchResultsWindow == null )
			{

				searchResults = new SearchResultsView();
				searchResults.ItemSelected += new RSSReader.Controls.SearchResultsView.ItemSelectedHandler(FeedView_FeedItemSelected);

				searchResultsWindow = new Toolwindow( new Guid( SEARCHRESULTS_TOOLWINDOW ), "RSS Reader Search Results", searchResults );
				searchResultsWindow.Visible = true;
				searchResultsWindow.Linkable = true;
				searchResultsWindow.IsFloating = false;
				searchResultsWindow.SetTabPicture( this.GetToolwindowIcon() );

			}

			searchResultsWindow.Caption = "Search Results for " + results.Keywords + " - " + results.Results.Count + " matches found";

			searchResults.SearchResults = results;
			searchResultsWindow.Activate();

		}

		private void FeedView_FeedItemChanged(object sender, RSSItem item)
		{
			tvwSubscriptions.UpdateFeedNode( feedView.Feed );
		}

		private void button_Click(object sender, EventArgs e)
		{

			if( this.subscriptionsWindow == null )
				InitExplorer();

			this.subscriptionsWindow.Visible = true;
			this.subscriptionsWindow.Activate();

		}

	}

}
